# Copyright (c) OpenMMLab. All rights reserved.
from .show_result import (show_multi_modality_result, show_result,
                          show_seg_result)

__all__ = ['show_result', 'show_seg_result', 'show_multi_modality_result']
